{-# LANGUAGE TypeFamilies #-}

import Data.Monoid
import Data.Foldable
import qualified Data.Set as Set
import qualified Data.Map as Map
import Control.Applicative

--import Control.Monad.Writer

studentList = [("Tom", 92),("Jerry", 43),
               ("Minnie", 76.5), ("Donald", 98), 
               ("Tom", 92), ("Micky", 76.3), 
               ("Shrek", 100), ("Peter-Pan", 63),
               ("Snoopy", 84), ("Popeye", 32)]
   
newtype SumPoints a = SumPoints { getSum :: a } deriving (Show)

instance (Num a) => Monoid ( SumPoints a ) where
   mempty = SumPoints 0
   mappend (SumPoints a) (SumPoints b) = SumPoints (a + b)
   
sumPoints :: (Num b) => (a, b) -> SumPoints b
sumPoints (a, b) = SumPoints b 

-- Example: get sum of all points
getSumPoints = foldMap sumPoints studentList
   
-- Student min points
data StudentMin a = StudentMin ([Char], a) | StudentMinEmpty deriving (Show)

instance (Ord a) => Monoid (StudentMin a) where
   mempty = StudentMinEmpty
   mappend StudentMinEmpty m = m
   mappend m StudentMinEmpty = m
   mappend (StudentMin (n1, p1)) ((StudentMin (n2, p2))) = if p1 <= p2 then StudentMin (n1, p1) else StudentMin (n2, p2)
   
stuMin :: ( Ord a) => ([Char], a) -> StudentMin a
stuMin (n, a) = StudentMin (n, a)

-- Example: get min points
getStuMin = foldMap stuMin studentList

-- Student max points
data StudentMax a = StudentMax ([Char], a) | StudentMaxEmpty deriving (Show)

instance (Ord a) => Monoid (StudentMax a) where
   mempty = StudentMaxEmpty
   mappend StudentMaxEmpty m = m
   mappend m StudentMaxEmpty = m
   mappend (StudentMax (n1, p1)) ((StudentMax (n2, p2))) = if p1 >= p2 then StudentMax (n1, p1) else StudentMax (n2, p2)
   
stuMax :: ( Ord a) => ([Char], a) -> StudentMax a
stuMax (n, a) = StudentMax (n, a)

-- Example: get max points
getStuMax = foldMap stuMax studentList
   
newtype Count = Count Int deriving (Show)

instance Monoid Count where
   mempty = Count 0
   mappend (Count a) (Count b) = Count (a + b)
   
count :: a -> Count
count _ = Count 1

-- Example: get the number of students
countStu = foldMap count studentList

-- Composition
  
compose2 :: (a -> b) -> (a -> c) -> a -> (b,c)
compose2 b c = (,) <$> b <*> c

compose3 :: (a -> b) -> (a -> c) -> (a -> d) -> a -> (b,c,d)
compose3 b c d = (,,) <$> b <*> c <*> d

-- Example: min points and max points
minmaxPoints = foldMap (compose2 stuMin stuMax) studentList

-- Example: count, min, max and sum
pointsInfo = foldMap (compose3  count (compose2 stuMin stuMax) sumPoints) studentList

-- define average

newtype Average a = Average (SumPoints a,Count) deriving (Show)

instance (Num a) => Monoid (Average a) where
  mempty = Average mempty
  mappend (Average m1) (Average m2) = Average (mappend m1 m2)
  
average :: (Num b) => (a, b) -> Average b
average a = Average (sumPoints a ,Count 1)

-- Example: average points
averagePoints = foldMap  average  studentList 

getAverage = let (Average (SumPoints s,Count n)) = averagePoints in s / fromIntegral n


-- Aggregation

-- define type class Aggregation
class (Monoid a) => Aggregation a where
  type AggResult a :: *
  aggResult :: a -> AggResult a
  
instance (Num a) => Aggregation (SumPoints a)  where
  type AggResult (SumPoints a) = a
  aggResult (SumPoints a) = a

instance (Ord a) => Aggregation (StudentMin a)  where
  type AggResult (StudentMin a) = ([Char], a)
  aggResult (StudentMin a) = a

instance (Ord a) => Aggregation (StudentMax a)  where
  type AggResult (StudentMax a) = ([Char], a)
  aggResult (StudentMax a) = a

instance Aggregation Count where
  type AggResult Count = Int
  aggResult (Count n) = n

instance (Aggregation a, Aggregation b) => Aggregation (a,b) where
  type AggResult (a,b) = (AggResult a, AggResult b)
  aggResult (a,b) = (aggResult a, aggResult b)

instance (Aggregation a, Aggregation b, Aggregation c) => Aggregation (a,b,c) where
  type AggResult (a,b,c) = (AggResult a, AggResult b, AggResult c)
  aggResult (a,b,c) = (aggResult a, aggResult b, aggResult c)
  
instance (Fractional a) => Aggregation (Average a) where
  type AggResult (Average a) = a
  aggResult (Average (SumPoints t,Count n)) = t/fromIntegral n
  
-- helper function aggResult (foldMap ...)

aggfoldMap :: (Foldable t, Aggregation a) => (v -> a) -> t v -> AggResult a
aggfoldMap f vs = aggResult (foldMap f vs)

-- Example: aggregation average
aggStuInfo = aggfoldMap (compose3 count (compose2 stuMin stuMax) average) studentList

-- Count unique
newtype UniqueCount a = UniqueCount (Set.Set a) deriving (Show)

instance (Ord a) => Monoid (UniqueCount a) where
   mempty = UniqueCount (Set.empty)
   mappend (UniqueCount set1) (UniqueCount set2) = UniqueCount (Set.union set1 set2)
   
instance (Ord a) => Aggregation (UniqueCount a) where
   type AggResult (UniqueCount a) =  Int
   aggResult (UniqueCount a) = Set.size a 
   
uCount :: (Ord a) => a -> UniqueCount a
uCount a = UniqueCount (Set.singleton a)

-- Example  count unique
getUCount = aggResult $ foldMap uCount studentList

getUCount' = aggResult $ foldMap (compose2 uCount count) studentList
